"""
********************************************************************************
compas_eve.memory
********************************************************************************

.. currentmodule:: compas_eve.memory


Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    InMemoryTransport

"""
from compas_eve.event_emitter import EventEmitterMixin
from compas_eve.core import Transport

__all__ = ["InMemoryTransport"]


class InMemoryTransport(Transport, EventEmitterMixin):
    """In-Memory transport is ideal for simple single-process apps and testing.

    It will only distribute messages within the same process, not across different processes."""

    def __init__(self, *args, **kwargs):
        super(InMemoryTransport, self).__init__(*args, **kwargs)
        self._local_callbacks = {}

    def on_ready(self, callback):
        """In-memory transport is always ready, it will immediately trigger the callback."""
        callback()

    def publish(self, topic, message):
        """Publish a message to a topic.

        Parameters
        ----------
        topic : :class:`Topic`
            Instance of the topic to publish to.
        message : :class:`Message`
            Instance of the message to publish.
        """
        event_key = "event:{}".format(topic.name)

        def _callback(**kwargs):
            self.emit(event_key, message)

        self.on_ready(_callback)

    def subscribe(self, topic, callback):
        """Subscribe to a topic.

        Every time a new message is received on the topic, the callback will be invoked.

        Parameters
        ----------
        topic : :class:`Topic`
            Instance of the topic to subscribe to.
        callback : function
            Callback to invoke whenever a new message arrives. The callback should
            receive only one `msg` argument, e.g. ``lambda msg: print(msg)``.

        Returns
        -------
        str
            Returns an identifier of the subscription.
        """
        event_key = "event:{}".format(topic.name)
        subscribe_id = "{}:{}".format(event_key, id(callback))

        def _callback(**kwargs):
            self.on(event_key, callback)

        self._local_callbacks[subscribe_id] = callback

        self.on_ready(_callback)

        return subscribe_id

    def unsubscribe_by_id(self, subscribe_id):
        """Unsubscribe from the specified topic based on the subscription id."""
        ev_type, topic_name, _callback_id = subscribe_id.split(":")
        event_key = "{}:{}".format(ev_type, topic_name)

        callback = self._local_callbacks[subscribe_id]
        self.off(event_key, callback)
        del self._local_callbacks[subscribe_id]

    def unsubscribe(self, topic):
        """Unsubscribe from the specified topic."""
        event_key = "event:{}".format(topic.name)
        self.remove_all_listeners(event_key)

    def advertise(self, topic):
        """Announce this code will publish messages to the specified topic.

        This call has no effect on the in-memory transport."""
        advertise_id = "advertise:{}:{}".format(topic.name, self.id_counter)
        # in-memory does not need anything here
        return advertise_id

    def unadvertise(self, topic):
        """Announce that this code will stop publishing messages to the specified topic.

        This call has no effect on the in-memory transport."""
        pass



# Python 2/3 compatibility import list
try:
    from collections import UserDict
except ImportError:
    from UserDict import UserDict

DEFAULT_TRANSPORT = None


def get_default_transport():
    """Retrieve the default transport implementation to be used system-wide.

    Returns
    -------
    :class:`~compas_eve.Transport`
        Instance of a transport class. By default, ``compas_eve`` uses
        :class:`~compas_eve.memory.InMemoryTransport`.
    """
    return DEFAULT_TRANSPORT


def set_default_transport(transport):
    """Assign a default transport implementation to be used system-wide.

    Parameters
    ----------
    transport : :class:`~compas_eve.Transport`
        Instance of a transport class. By default, ``compas_eve`` uses
        :class:`~compas_eve.memory.InMemoryTransport`.
    """
    global DEFAULT_TRANSPORT
    DEFAULT_TRANSPORT = transport


class Transport(object):
    """Defines the base interface for different transport implementations."""

    def __init__(self, *args, **kwargs):
        super(Transport, self).__init__(*args, **kwargs)
        self._id_counter = 0

    @property
    def id_counter(self):
        """Generate an auto-incremental ID starting from 1."""
        self._id_counter += 1
        return self._id_counter

    def publish(self, topic, message):
        pass

    def subscribe(self, topic, callback):
        pass

    def unsubscribe(self, topic):
        pass

    def advertise(self, topic):
        pass

    def unadvertise(self, topic):
        pass


class Message(UserDict):
    """Message objects used for publishing and subscribing to/from topics.

    A message is fundamentally a dictionary and behaves as one."""

    def __str__(self):
        return str(self.data)

    def __getattr__(self, name):
        return self.__dict__["data"][name]

    @classmethod
    def parse(cls, value):
        instance = cls()
        instance.update(value)
        return instance


class Topic(object):
    """A topic is like a mailbox where messages can be sent and received.

    Topics are described by a name, a type of message they accept, and a
    set of options.

    Attributes
    ----------
    name : str
        Name of the topic.
    message_type : type
        Class defining the message structure. Use :class:`Message` for
        a generic, non-typed checked message implementation.
    options : dict
        A dictionary of options.
    """

    # TODO: Add documentation/examples of possible options

    def __init__(self, name, message_type, **options):
        self.name = name
        self.message_type = message_type
        self.options = options


class Publisher(object):
    """Publisher interface."""

    def __init__(self, topic, transport=None):
        self.topic = topic
        self.transport = transport or get_default_transport()
        self._advertise_id = None

    @property
    def is_advertised(self):
        """Indicate if the publisher has announce its topic as advertised or not.

        Returns:
            bool: True if advertised as publisher of this topic, False otherwise.
        """
        return self._advertise_id is not None

    def message_published(self, message):
        """Handler called when a message has been published."""
        pass

    def publish(self, message):
        """Publish a message to a topic."""
        # TODO: check if message type matches self.topic.message_type declared
        if not self.is_advertised:
            self.advertise()

        self.transport.publish(self.topic, message)
        self.message_published(message)

    def advertise(self):
        if self.is_advertised:
            return

        self._advertise_id = self.transport.advertise(self.topic)

    def unadvertise(self):
        if not self.is_advertised:
            return

        self.transport.unadvertise(self.topic)
        self._advertise_id = None


class Subscriber(object):
    """Subscriber interface."""

    def __init__(self, topic, callback=None, transport=None):
        self.transport = transport or get_default_transport()
        self.topic = topic
        self._subscribe_id = None
        self._callback = callback

    def message_received(self, message):
        """Handler called whenever a new message is received.

        By default, this implementation will simply invoke the callback
        used on init (if any), but sub-classes can override this behavior."""
        if self._callback:
            self._callback(message)

    @property
    def is_subscribed(self):
        """Indicate if the instace is currently subscribed to its topic or not."""
        return self._subscribe_id is not None

    def subscribe(self):
        if self._subscribe_id:
            return

        self._subscribe_id = self.transport.subscribe(self.topic, self.message_received)

    def unsubscribe(self):
        """Unregister the subscriber from its topic."""
        if not self._subscribe_id:
            return

        self.transport.unsubscribe_by_id(self._subscribe_id)
        self._subscribe_id = None



from compas.data import json_dumps
from compas.data import json_loads

from ..core import Transport
from ..event_emitter import EventEmitterMixin

import paho.mqtt.client as mqtt


class MqttTransport(Transport, EventEmitterMixin):
    """MQTT transport allows sending and receiving messages using an MQTT broker.

    Parameters
    ----------
    host : str
        Host name for the MQTT broker, e.g. ``broker.hivemq.com`` or ``localhost`` if
        you are running a local broker on your machine.
    port : int
        MQTT broker port, defaults to ``1883``.
    """

    def __init__(self, host, port=1883, *args, **kwargs):
        super(MqttTransport, self).__init__(*args, **kwargs)
        self.host = host
        self.port = port
        self._is_connected = False
        self._local_callbacks = {}
        self.client = mqtt.Client()  # todo: generate client_id
        self.client.on_connect = self._on_connect
        self.client.connect(self.host, self.port)
        self.client.loop_start()

    def close(self):
        """Close the connection to the MQTT broker."""
        self.client.loop_stop()

    def _on_connect(self, client, userdata, flags, rc):
        self._is_connected = True
        self.emit("ready")

    def on_ready(self, callback):
        """Allows to hook-up to the event triggered when the connection to MQTT broker is ready.

        Parameters
        ----------
        callback : function
            Function to invoke when the connection is established.
        """
        if self._is_connected:
            callback()
        else:
            self.once("ready", callback)

    def publish(self, topic, message):
        """Publish a message to a topic.

        Parameters
        ----------
        topic : :class:`Topic`
            Instance of the topic to publish to.
        message : :class:`Message`
            Instance of the message to publish.
        """

        def _callback(**kwargs):
            # TODO: can we avoid the additional cast to dict?
            json_message = json_dumps(dict(message))
            self.client.publish(topic.name, json_message)

        self.on_ready(_callback)

    def subscribe(self, topic, callback):
        """Subscribe to a topic.

        Every time a new message is received on the topic, the callback will be invoked.

        Parameters
        ----------
        topic : :class:`Topic`
            Instance of the topic to subscribe to.
        callback : function
            Callback to invoke whenever a new message arrives. The callback should
            receive only one `msg` argument, e.g. ``lambda msg: print(msg)``.

        Returns
        -------
        str
            Returns an identifier of the subscription.
        """
        event_key = "event:{}".format(topic.name)
        subscribe_id = "{}:{}".format(event_key, id(callback))

        def _local_callback(msg):
            msg = topic.message_type.parse(json_loads(msg.payload.decode()))
            callback(msg)

        def _subscribe_callback(**kwargs):
            self.client.subscribe(topic.name)

            # We only really need to hook up once per client
            if not self.client.on_message:
                self.client.on_message = self._on_message

            self.on(event_key, _local_callback)

        self._local_callbacks[subscribe_id] = _local_callback

        self.on_ready(_subscribe_callback)

        return subscribe_id

    def _on_message(self, client, userdata, msg):
        event_key = "event:{}".format(msg.topic)
        self.emit(event_key, msg)

    def advertise(self, topic):
        """Announce this code will publish messages to the specified topic.

        This call has no effect on this transport implementation.

        Parameters
        ----------
        topic : :class:`Topic`
            Instance of the topic to advertise.

        Returns
        -------
        str
            Advertising identifier.
        """

        advertise_id = "advertise:{}:{}".format(topic.name, self.id_counter)
        # mqtt does not need anything here
        return advertise_id

    def unadvertise(self, topic):
        """Announce that this code will stop publishing messages to the specified topic.

        This call has no effect on this transport implementation.

        Parameters
        ----------
        topic : :class:`Topic`
            Instance of the topic to stop publishing messages to.
        """
        pass

    def unsubscribe_by_id(self, subscribe_id):
        """Unsubscribe from the specified topic based on the subscription id.

        Parameters
        ----------
        subscribe_id : str
            Identifier of the subscription.
        """
        ev_type, topic_name, _callback_id = subscribe_id.split(":")
        event_key = "{}:{}".format(ev_type, topic_name)

        callback = self._local_callbacks[subscribe_id]
        self.off(event_key, callback)
        self.client.unsubscribe(topic_name)

        del self._local_callbacks[subscribe_id]

    def unsubscribe(self, topic):
        """Unsubscribe from the specified topic.

        Parameters
        ----------
        topic : :class:`Topic`
            Instance of the topic to unsubscribe from.
        """
        self.client.unsubscribe(topic.name)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# flake8: noqa
from __future__ import absolute_import
from __future__ import print_function

import io
from os import path

from setuptools import setup
from setuptools.command.develop import develop
from setuptools.command.install import install


here = path.abspath(path.dirname(__file__))


def read(*names, **kwargs):
    return io.open(path.join(here, *names), encoding=kwargs.get("encoding", "utf8")).read()


long_description = read("README.md")
requirements = read("requirements.txt").split("\n")
optional_requirements = {}

setup(
    name="compas_eve",
    version="0.3.0",
    description="COMPAS Event Extensions: adds event-based communication infrastructure to the COMPAS framework.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/compas-dev/compas_eve",
    author="Gonzalo Casas",
    author_email="casas@arch.ethz.ch",
    license="MIT license",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    keywords=[],
    project_urls={},
    packages=["compas_eve"],
    package_dir={"": "src"},
    package_data={},
    data_files=[],
    include_package_data=True,
    zip_safe=False,
    install_requires=requirements,
    python_requires=">=3.8",
    extras_require=optional_requirements,
    entry_points={
        "console_scripts": [],
    },
    ext_modules=[],
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import glob
import os

import compas
import compas.plugins
from compas_ghpython.components import install_userobjects
from compas_ghpython.components import uninstall_userobjects


@compas.plugins.plugin(category="install")
def installable_rhino_packages():
    return ["compas_eve"]


@compas.plugins.plugin(category="install")
def after_rhino_install(installed_packages):
    project = "compas_eve"
    if project not in installed_packages:
        return []

    srcdir = os.path.join(os.path.dirname(__file__), "..", "ghpython", "components", "ghuser")
    installed_objects = install_userobjects(srcdir)
    msg = "Installed {} GH User Objects".format(len(installed_objects))

    return [
        (
            project,
            msg,
            True,
        )
    ]


@compas.plugins.plugin(category="install")
def after_rhino_uninstall(uninstalled_packages):
    project = "compas_eve"
    if project not in uninstalled_packages:
        return []

    srcdir = os.path.join(os.path.dirname(__file__), "..", "ghpython", "components", "ghuser")
    userobjects = [os.path.basename(ghuser) for ghuser in glob.glob(os.path.join(srcdir, "*.ghuser"))]
    uninstalled_objects = uninstall_userobjects(userobjects)

    uninstall_errors = [uo[0] for uo in uninstalled_objects if not uo[1]]
    error_msg = "" if not uninstall_errors else "and {} failed to uninstall".format(len(uninstall_errors))
    msg = "Uninstalled {} GH User Objects {}".format(len(uninstalled_objects), error_msg)

    return [
        (
            project,
            msg,
            True,
        )
    ]
